% Matlab code implementing the Pruning Model Selection scheme for Hidden 
% Markov Models as described in the paper:
%
% M. Bicego, V. Murino, M.A.T. Figueiredo: "A sequential pruning strategy 
% for the selection of the number of states in Hidden Markov Models". 
% Pattern Recognition Letters , vol. 24(9-10), pp. 1395-1407, (2003)
%
% These routines are based on the HMM library provided by Kevin Murphy (you should download and install it)
% ---------------
% Hidden Markov Model (HMM) Toolbox written by Kevin Murphy (1998).
% See http://www.ai.mit.edu/~murphyk/Software/hmm.html for details.
% ---------------
%
% The model selection scheme is for continuous Gaussian HMM (1 Gaussian 
% per state) with the BIC (Bayesian Information Criterion) criterion.
% 
% Note: For continuous HMM, we used only one Gaussian per state since 
% it has been shown in the paper that using K states with M-components 
% mixture per state is equivalent to use a KM states model with 1 gaussian 
% per state
% 
clear all
close all


%% Generating sequences: This is the example provided in the paper 
% (synthetic data, model "a")
load('SyntethicExampleA.mat');
% % Here is the code used to genenerate it
% T = 400; % sequence length
% O = 1; % dimensionality
% nex = 5; % number of sequences
% origtransmat = ones(4)./4;    % transition matrix to generate 
% origprior = ones(1,4)*0.25;  % init state matrix to generate 
% m = [-2 0 2 4];              % 1D gaussian per state: mean  
% s = [0.5 0.5 0.5 0.5];         % 1D gaussian per state: std
% M = 1;  % 1 gaussian per mixture
% for i=1:nex
%     seqs(:,:,i) = genera_seq_gauss(origtransmat,m,s,origprior,T)';
% end
% save('SyntethicExampleC.mat','seqs');


%%  This is another example provided in the paper (synthetic data, model c)
% load('SyntethicExampleC.mat'); % uncomment if you want to use this
% T = 400; % sequence length
% O = 1; % dimensionality
% nex = 5; % number of sequences
% origtransmat = [0.85 0.05 0.05 0.05; 
%     0.05 0.85 0.05 0.05;
%     0.05 0.05 0.85 0.05;
%     0.05 0.05 0.05 0.85];    % transition matrix to generate 
% origprior = ones(1,4)*0.25;  % init state matrix to generate 
% m = [0 0 -4 4];              % 1D gaussian per state: mean  
% s = [0.2 2 0.5 0.5];         % 1D gaussian per state: std
% M = 1;  % 1 gaussian per mixture
% for i=1:nex
%     seqs(:,:,i) = genera_seq_gauss(origtransmat,m,s,origprior,T)';
% end
% save('SyntethicExampleC.mat','seqs');

%% Training of HMM
K_min = 2;
K_max = 10;
Ktrue = 4;
O = 1;
M = 1;

% % Learning standard HMM with Ktrue states
% initial guess of parameters
prior0 = normalise(rand(Ktrue,1));
transmat0 = mk_stochastic(rand(Ktrue,Ktrue));
[mu0, Sigma0] = mixgauss_init(Ktrue, seqs, 'full'); % initialization via gmm clustering
mu0 = reshape(mu0, [O Ktrue M]);
Sigma0 = reshape(Sigma0, [O O Ktrue M]);
mixmat0 = mk_stochastic(rand(Ktrue,M));
[LL, prior1, transmat1, mu1, Sigma1, mixmat1] = mhmm_em(seqs, prior0, transmat0, mu0, Sigma0, mixmat0,'max_iter', 100);
LL(end)

% Automatically learning the number of states
% via pruning model selection
% initial guess of parameters: model with K_max states
prior0 = normalise(rand(K_max,1));
transmat0 = mk_stochastic(rand(K_max,K_max));
[mu0, Sigma0] = mixgauss_init(K_max, seqs, 'full'); % initialization via gmm clustering
mu0 = reshape(mu0, [O K_max M]);
Sigma0 = reshape(Sigma0, [O O K_max M]);
mixmat0 = mk_stochastic(rand(K_max,M));
[LL, prior1, transmat1, mu1, Sigma1, mixmat1] = PruningMSmhmm_em(seqs, prior0, transmat0, mu0, Sigma0, mixmat0,K_min);
fprintf('Best number of states %d\n',length(prior1));
LL(end)

